"""Object stored in the file."""
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional

from goodreads_export.templates import FileTemplate

if TYPE_CHECKING:
    from goodreads_export.library import Library


class DataFile:
    """Object stored in the file."""

    library: "Library"
    folder: Optional[Path]

    _file_name: Optional[Path]
    _content: Optional[str]

    def __init__(
        self,
        *,
        library: "Library",
        folder: Optional[Path] = None,
        file_name: Optional[Path] = None,
        content: Optional[str] = None,
    ) -> None:
        """Set fields from args."""
        self.library = library
        self.folder = folder
        self._file_name = file_name
        self._content = content

    def _get_template(self) -> FileTemplate:
        """Template."""
        raise NotImplementedError()

    def _get_template_context(self) -> Dict[str, Any]:
        """Return template context."""
        raise NotImplementedError()

    @property
    def file_name(self) -> Path:
        """Markdown file name.

        Automatically generate file name from book's fields if not assigned.
        """
        if self._file_name is None:
            self._file_name = self._get_template().render_file_name(self._get_template_context())
        return self._file_name

    @file_name.setter
    def file_name(self, file_name: Path) -> None:
        """Set file_name."""
        self._file_name = file_name

    @property
    def file_link(self) -> str:
        """Return file link."""
        return self._get_template().render_file_link({"file_name": self.file_name})

    def render_body(self) -> str:
        """Return rendered body."""
        raise NotImplementedError()

    def parse(self) -> None:
        """Parse file content."""
        raise NotImplementedError()

    @property
    def content(self) -> str:
        """File content.

        Automatically generate content from object's fields if not assigned.
        """
        if self._content is None:
            self._content = self.render_body()
        return self._content

    @content.setter
    def content(self, content: str) -> None:
        """Set content and parse it."""
        self._content = content
        self.parse()

    def delete_file(self) -> None:
        """Delete the series file."""
        if self.path.exists():
            os.remove(self.path)

    @property
    def path(self) -> Path:
        """Return file path."""
        assert self.folder is not None
        return self.folder / self.file_name

    def write(self) -> None:
        """Write file to path."""
        self.path.write_text(self.content, encoding="utf8")



"""Book's file."""
import os
import urllib.parse
from pathlib import Path
from typing import Any, Dict, List, Optional

from goodreads_export.authored_file import AuthoredFile
from goodreads_export.series_file import SeriesFile
from goodreads_export.templates import BookTemplate


class BookFile(AuthoredFile):  # pylint: disable=too-many-instance-attributes
    """Book's markdown file.

    On init extract fields from `content` - override other parameters.
    To re-parse the `content` call `parse()`.
    `render()` generate `content` from fields.
    """

    title: Optional[str]
    book_id: Optional[str]
    tags: List[str]
    rating: Optional[int]
    isbn: Optional[int]
    isbn13: Optional[int]
    review: Optional[str]
    series_titles: List[str]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        title: Optional[str] = None,
        book_id: Optional[str] = None,
        tags: Optional[List[str]] = None,
        rating: Optional[int] = None,
        isbn: Optional[int] = None,
        isbn13: Optional[int] = None,
        review: Optional[str] = None,
        series_titles: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Init."""
        super().__init__(**kwargs)
        self.title = title
        self.book_id = book_id
        self.tags = tags or []
        self.rating = rating
        self.isbn = isbn
        self.isbn13 = isbn13
        self.review = review
        self.series_titles = series_titles or []
        self.parse()

    def _get_template(self) -> BookTemplate:
        """Template."""
        return self.library.templates.book

    def _get_template_context(self) -> Dict[str, Any]:
        """Template context."""
        return {
            "book": self,
            "author": self.author,
            "urlencode": urllib.parse.urlencode,
        }

    def parse(self) -> None:
        """Parse file content."""
        if self._content is None:
            return
        self.book_id = None
        self.title = None
        self.author = None
        self.series_titles = []
        if book_regex := self._get_template().goodreads_link_regexes.choose_regex(self._content):
            link_match = book_regex.compiled.search(self._content)
            assert link_match is not None  # to please mypy
            self.book_id = link_match[book_regex.book_id_group]
            self.title = link_match[book_regex.title_group]
            self.author = self.library.get_author(link_match[book_regex.author_group])
        if series_regex := self._get_template().series_regexes.choose_regex(self._content):
            self.series_titles = [
                series_match[series_regex.series_group]
                for series_match in series_regex.compiled.finditer(self._content)
            ]

    def render_body(self) -> str:
        """Render file body."""
        assert self.tags is not None  # to please mypy
        if "#book/book" not in self.tags:
            self.tags.append("#book/book")
        if self.rating is not None and self.rating > 0:
            rating_tag = f"#book/rating{self.rating}"
            if rating_tag not in self.tags:
                self.tags.append(rating_tag)
        return self._get_template().render_body(self._get_template_context())

    def write(self) -> None:
        """Write markdown file to path.

        If the file with `author - title.md` already exists, append book ID to the file name.
        Even if this file also exists that does not matter because this it the book
        with the same ID.
        """
        assert self.folder is not None  # to please mypy
        assert self.book_id is not None  # to please mypy
        if (self.folder / self.file_name).exists() and self.book_id not in str(self.file_name):
            self.file_name = Path(
                f"{self.file_name.with_suffix('')} - {self.book_id}{self.file_name.suffix}"
            )
        super().write()

    def delete_series_files(self) -> Dict[str, Path]:
        """Delete series files for review.

        Return deleted series files {series name: series file path}
        """
        deleted_series_files = {}
        assert self.series_titles is not None
        for series in self.series_titles:
            series_file_path = self.folder / self.series_file_name(series)  # type: ignore  # pylint: disable=no-member
            if series_file_path.exists():
                os.remove(series_file_path)
                deleted_series_files[series] = series_file_path
        return deleted_series_files

    def create_series_files(self) -> Dict[str, Path]:
        """Create series files if they do not exist.

        Do not change already existed files.
        Return created series files {series name: series file path}
        """
        created_series_files = {}
        for series in self.series:
            if not series.path.is_file():
                assert series.title is not None  # to please mypy
                series.write()
                created_series_files[series.title] = series.path
        return created_series_files

    def check(self) -> bool:
        """Check regexps for the template.

        Create file from fields and after that parse it and compare parsed values with the initial fields
        """
        book_id = self.book_id
        title = self.title
        assert self.author is not None  # to please mypy
        author_name = self.author.name
        series_titles = self.series_titles
        self.content = self.render_body()
        is_book_id_parsed = self.book_id == book_id
        is_title_parsed = self.title == title
        is_author_parsed = self.author.name == author_name
        is_series_parsed = self.series_titles == series_titles
        # assert book_file.series == ['']
        if not is_author_parsed:
            print(f"Author name {author_name} is not parsed from content\n{self.content}")
            print(f"using the pattern\n{self._get_template().goodreads_link_regexes[0].regex}")
        if not is_book_id_parsed:
            print(f"Book ID {book_id} is not parsed from content\n{self.content}")
            print(f"using the pattern\n{self._get_template().goodreads_link_regexes[0].regex}")
        if not is_title_parsed:
            print(f"Book title {title} is not parsed from content\n{self.content}")
            print(f"using the pattern\n{self._get_template().goodreads_link_regexes[0].regex}")
        if not is_series_parsed:
            print(f"Series {series_titles[0]} is not parsed from content\n{self.content}")
            print(f"using the pattern\n{self._get_template().series_regexes[0].regex}")
        return is_book_id_parsed and is_title_parsed and is_author_parsed and is_series_parsed

    @property
    def series(self) -> List[SeriesFile]:
        """List of series objects constructed from series_names."""
        return [
            SeriesFile(library=self.library, folder=self.folder, title=title, author=self.author)
            for title in self.series_titles
        ]



"""Series file."""
import urllib.parse
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from goodreads_export.authored_file import AuthoredFile
from goodreads_export.templates import SeriesTemplate


class SeriesFile(AuthoredFile):
    """Series' file."""

    title: Optional[str]

    def __init__(self, *, title: Optional[str] = None, **kwargs: Any) -> None:
        """Extract fields from content."""
        super().__init__(**kwargs)
        self.title = title
        self.parse()

    def _get_template(self) -> SeriesTemplate:
        """Template."""
        return self.library.templates.series

    def _get_template_context(self) -> Dict[str, Any]:
        """Return template context for series."""
        return {
            "series": self,
            "author": self.author,
            "urlencode": urllib.parse.urlencode,
        }

    def parse(self) -> None:
        """Parse file content."""
        if self._content is not None:
            self.title = None
            self.author = None
            if regex := self._get_template().content_regexes.choose_regex(self._content):
                match = regex.compiled.search(self._content)
                assert match  # to make mypy happy
                self.title = match[regex.title_group]
                self.author = self.library.get_author(match[regex.author_group])

    def is_file_name(self, file_name: Union[str, Path]) -> bool:
        """Return True if file name if indicate this is series description file."""
        return self._get_template().file_name_regexes.choose_regex(str(file_name)) is not None

    def render_body(self) -> str:
        """Render series body."""
        return self._get_template().render_body(self._get_template_context())

    def check(self) -> bool:
        """Check regexps for the template.

        Create file from fields and after that parse it and compare parsed values with the initial fields
        """
        title = self.title
        assert self.author is not None  # to please mypy
        author_name = self.author.name
        self.content = self.render_body()
        is_title_parsed = self.title == title
        is_author_parsed = self.author.name == author_name
        if not is_title_parsed:
            print(f"Series title {title} is not parsed from content\n{self.content}")
            print(f"using the pattern\n{self._get_template().content_regexes[0].regex}")
        if not is_author_parsed:
            print(f"Author name {author_name} is not parsed from content\n{self.content}")
            print(f"using the pattern\n{self._get_template().content_regexes[0].regex}")
        series_file_name = self.file_name
        assert series_file_name is not None  # to please mypy
        is_file_name = self.is_file_name(series_file_name)
        if not is_file_name:
            print(f"Series file name {series_file_name} is not recognized")
            print(f"using the pattern\n{self._get_template().file_name_regexes[0].regex}")
        return is_title_parsed and is_author_parsed and is_file_name


class SeriesList(List[SeriesFile]):
    """List of SeriesFile objects."""

    def by_title(self) -> Optional[SeriesFile]:
        """Find 1st series with the title."""
        return None



"""Library object with author."""
from abc import ABC
from typing import TYPE_CHECKING, Any, Optional

from goodreads_export.data_file import DataFile

if TYPE_CHECKING:
    from goodreads_export.library import AuthorFile


class AuthoredFile(DataFile, ABC):
    """Authored data file."""

    author: Optional["AuthorFile"]

    def __init__(self, *, author: Optional["AuthorFile"] = None, **kwargs: Any) -> None:
        """Set fields from args."""
        super().__init__(**kwargs)
        self.author = author

    def rename_author(self, new_author: str) -> None:
        """Rename author.

        We do not re-render the file fully to keep intact possible user changes in it.
        """
        self.delete_file()
        assert self.author is not None  # to please mypy
        old_author_link = self.author.file_link
        self.author = self.library.get_author(new_author)
        self._file_name = None  # to force re-rendering
        self.content = self.content.replace(old_author_link, self.author.file_link)
        self.write()



"""Library of ooks."""
import os
from pathlib import Path
from typing import Dict, Optional

from goodreads_export.author_file import AuthorFile
from goodreads_export.book_file import BookFile
from goodreads_export.goodreads_book import Book, GoodreadsBooks
from goodreads_export.log import Log
from goodreads_export.series_file import SeriesFile
from goodreads_export.stat import Stat
from goodreads_export.templates import TemplateSet, TemplatesLoader

SUBFOLDERS = {
    "toread": "toread",  # for books without review and rating - supposedly this is from to-read
    "reviews": "reviews",  # all other books
    "authors": "authors",  # book authors
}

BOOKS_SUBFOLDERS = [SUBFOLDERS["reviews"], SUBFOLDERS["toread"]]


class Library:
    """Books and authors."""

    stat = Stat()

    def __init__(
        self,
        folder: Optional[Path] = None,
        log: Optional[Log] = None,
        templates: Optional[TemplateSet] = None,
    ) -> None:
        """Initialize."""
        self.folder = folder
        self.log = log or Log()
        self.templates = templates or TemplatesLoader().load_builtin()

        self.books: Dict[str, BookFile] = {}
        self.authors: Dict[str, AuthorFile] = {}
        self.primary_authors: Dict[str, AuthorFile] = {}
        if folder is not None:
            self.authors = self.load_authors(folder / SUBFOLDERS["authors"])
            for books_subfolder in BOOKS_SUBFOLDERS:
                self.load_series(folder / books_subfolder, self.authors)
            for books_subfolder in BOOKS_SUBFOLDERS:
                self.books |= self.load_books(folder / books_subfolder, self.authors)

    def merge_author_names(self) -> None:
        """Replace all known versions of author names (translations, misspellings) with one primary name.

        All author name versions should be listed as links in one author file -
        just copy them from other author files to that `primary` author file.
        Fist name in the link will be used as primary name.

        Reviews will be relinked to the primary name.
        Author files with non-primary names will be deleted.
        Also recreate series files with primary author name in the file names.
        """
        for primary_author in [
            author
            for author in self.authors.values()
            if len(set(author.names) - {author.name}) > 0
        ]:
            for author_name in primary_author.names:
                if (
                    author_name in self.authors
                    and (author := self.authors[author_name]) != primary_author
                ):
                    self.log.debug(
                        f"Author {primary_author.name} has synonim {author.name} to merge"
                    )
                    self.stat.authors_renamed += 1
                    primary_author.merge(self.authors[author_name])
                    self.authors[author_name] = primary_author

    def dump(self, books: GoodreadsBooks) -> None:
        """Save books and authors as md-files."""
        for subfolder in SUBFOLDERS.values():
            assert self.folder is not None  # for mypy
            os.makedirs(self.folder / subfolder, exist_ok=True)

        reviews_bar_title = "Review"
        authors_bar_title = "Author"
        self.log.open_progress(reviews_bar_title, "books", len(books))
        self.log.open_progress(authors_bar_title, "authors", bar_format="{desc}: {n_fmt}")

        for book in books:
            self.log.progress(reviews_bar_title)
            self.log.progress_description(reviews_bar_title, f"{book.title}")
            if self.stat.unique_author(book.author):
                self.log.progress(authors_bar_title)
            if book.author in self.authors and book.author != (
                primary_author := self.authors[book.author].name
            ):
                self.log.progress_description(
                    authors_bar_title, f"Author {book.author} changed to {primary_author}"
                )
                book.author = primary_author  # use the same author name for all synonyms

            if book.book_id not in self.books:
                if book.author not in self.authors:
                    self.authors[book.author] = self.create_author_file(book)
                    self.stat.authors_added += 1
                    self.log.progress_description(authors_bar_title, f"Added author {book.author}")
                added_file_path = self.create_book_file(book)
                # we know there was no file with this book ID so we added it for sure
                self.stat.books_added += 1
                self.log.debug(f"Added book {book.title}, {added_file_path} ")
        self.log.close_progress()

    def create_book_file(self, book: Book) -> str:
        """Create book file.

        Return the filename.
        """
        if book.review == "" and book.rating == 0:
            subfolder = SUBFOLDERS["toread"]
        else:
            subfolder = SUBFOLDERS["reviews"]
        assert self.folder is not None  # for mypy
        book_file = BookFile(
            library=self,
            title=book.title,
            folder=self.folder / subfolder,
            tags=book.tags,
            author=self.get_author(book.author),
            book_id=book.book_id,
            rating=book.rating,
            isbn=book.isbn,
            isbn13=book.isbn13,
            series_titles=book.series,
            review=book.review,
        )
        book_file.create_series_files()
        book_file.write()
        return os.path.join(subfolder, book_file.file_name)

    def create_author_file(self, book: Book) -> AuthorFile:
        """Create author file if id does not exist.

        Do not change already existed file.

        Return True if author file was added, False otherwise
        """
        assert self.folder is not None  # for mypy
        author_file = AuthorFile(
            library=self,
            name=book.author,
            folder=self.folder / SUBFOLDERS["authors"],
        )
        if not author_file.path.is_file():
            author_file.write()
        return author_file

    def get_author(self, name: str) -> AuthorFile:
        """Get author file."""
        if name not in self.authors:
            if self.folder is None:
                return AuthorFile(library=self, name=name)
            self.log.info(f"Creating author '{name}' ")
            self.authors[name] = AuthorFile(
                library=self, name=name, folder=self.folder / SUBFOLDERS["authors"]
            )
            self.authors[name].write()
        return self.authors[name]

    def load_series(self, folder: Path, authors: Dict[str, AuthorFile]) -> None:
        """Load existed series.

        Add them to authors.
        Could add series with the same title to the same author if they are in different files.
        """
        dummy_author = AuthorFile(library=Library(), name="author", folder=folder)
        dummy_series = SeriesFile(library=Library(), author=dummy_author, title="title")
        for file_name in folder.glob(f"*{dummy_series.file_name.suffix}"):
            if dummy_series.is_file_name(file_name):
                series = SeriesFile(
                    library=self,
                    folder=folder,
                    file_name=Path(file_name.name),
                    content=file_name.read_text(encoding="utf8"),
                )
                if series.author is None:
                    self.log.info(f"Series file {file_name} has no author name")
                    continue
                if series.author.name not in authors:
                    self.log.info(f"Series file {file_name} has author without author file")
                    continue
                authors[series.author.name].series.append(series)
                self.stat.series_added += 1

    def load_books(self, folder: Path, authors: Dict[str, AuthorFile]) -> Dict[str, BookFile]:
        """Load existed books.

        Look for goodreads book ID inside files.
        Return {id: BookFile} for files with book ID, ignore other files.
        This way we ignore "- series" files and unknown files.
        """
        books: Dict[str, BookFile] = {}
        dummy_author = AuthorFile(library=Library(), name="author", folder=folder)
        dummy_book = BookFile(library=Library(), author=dummy_author, title="title")
        dummy_series = SeriesFile(library=Library(), author=dummy_author, title="title")
        for file_name in folder.glob(f"*{dummy_book.file_name.suffix}"):
            book = BookFile(  # also create author file if not yet existed
                library=self,
                folder=folder,
                file_name=Path(file_name.name),
                content=file_name.read_text(encoding="utf8"),
            )
            if book.book_id is None or book.author is None:
                if not dummy_series.is_file_name(file_name):
                    self.stat.skipped_unknown_files += 1
            elif book.book_id in books:
                raise ValueError(
                    f"Duplicate book ID {book.book_id} in {file_name} "
                    f"and {books[book.book_id].file_name}"
                )
            else:
                books[book.book_id] = book
                authors[book.author.name].books.append(book)
        return books

    def load_authors(self, folder: Path) -> Dict[str, AuthorFile]:
        """Load existed authors.

        Return loaded authors
        """
        authors: Dict[str, AuthorFile] = {}
        dummy_author = AuthorFile(library=Library(), name="author", folder=folder)
        for file_name in folder.glob(f"*{dummy_author.file_name.suffix}"):
            author = AuthorFile(
                library=self,
                folder=folder,
                file_name=Path(file_name.name),
                name=file_name.stem,  # will be replaced by parsing file content
                content=file_name.read_text(encoding="utf8"),
            )
            if author.names:  # parse succeeded
                authors[author.name] = author  # primary name is always point to primary file
                for name in author.names:
                    if name not in authors:  # do not overwrite if pointed to primary file
                        authors[name] = author
            else:
                self.stat.skipped_unknown_files += 1
        return authors

    def check_templates(self) -> None:
        """Check templates and regexes."""
        author_name = "Jules Verne"
        book_id = "54479"
        title = "Around the World in Eighty Days"
        tags = ["#adventure", "#classics", "#fiction", "#novel", "#travel"]
        series_titles = ["Voyages extraordinaires"]
        review = "This is a review\nin two lines"
        author = AuthorFile(library=self, name=author_name)
        author.check()
        BookFile(
            library=self,
            book_id=book_id,
            title=title,
            author=author,
            tags=tags,
            series_titles=series_titles,
            review=review,
            rating=5,
        ).check()
        title = "title"
        SeriesFile(library=self, title=title, author=author).check()



"""Create md-files from https://www.goodreads.com/ CSV export."""
import os
import shutil
import sys
from pathlib import Path
from typing import Optional

import click

from goodreads_export.goodreads_book import GoodreadsBooks
from goodreads_export.library import Library
from goodreads_export.log import Log
from goodreads_export.templates import DEFAULT_BUILTIN_TEMPLATE, TemplateSet, TemplatesLoader
from goodreads_export.version import VERSION

GOODREAD_EXPORT_FILE_NAME = "goodreads_library_export.csv"
DEFAULT_TEMPLATES_FOLDER = "./templates"

VERBOSE_OPTION = click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="""Increase verbosity.""",
    nargs=1,
)

TEMPLATES_FOLDER_OPTION = click.option(
    "--templates-folder",
    "-t",
    "templates_folder",
    default=None,
    type=click.Path(path_type=Path),
    help=f"""Folder with templates. If not absolute it's relative to `BOOKS_FOLDER`.
If not specified, look for `{DEFAULT_TEMPLATES_FOLDER}` in `BOOKS_FOLDER`.
If not found use built-in templates, see `--builtin-name`.""",
    nargs=1,
)

BUILTIN_TEMPLATES_NAME_OPTION = click.option(
    "--builtin-name",
    "-b",
    "builtin_name",
    default=None,
    help=f"""Name of the built-in template. Use `{DEFAULT_BUILTIN_TEMPLATE}` if not specified.""",
    nargs=1,
)


BOOKS_FOLDER_OPTION = click.argument(
    "books_folder",
    type=click.Path(exists=True, path_type=Path),
    nargs=1,
)


BOOKS_FOLDER_OPTIONAL_OPTION = click.argument(
    "books_folder",
    required=False,
    type=click.Path(exists=True, path_type=Path),
    nargs=1,
)


def merge_authors(log: Log, books_folder: Path, templates: TemplateSet) -> Library:
    """Merge authors."""
    library = load_library(log=log, books_folder=books_folder, templates=templates)
    library.merge_author_names()
    return library


def load_library(log: Log, books_folder: Path, templates: TemplateSet) -> Library:
    """Load library."""
    library = Library(folder=books_folder, log=log, templates=templates)
    log.start(f"Reading existing files from {books_folder}")
    print(
        f" loaded {len(library.books)} books, {len(library.authors)} authors, "
        f"skipped {library.stat.skipped_unknown_files} unknown files"
        f" and {library.stat.series_added} series files.",
    )
    return library


def load_templates(
    log: Log,
    books_folder: Optional[Path],
    templates_folder: Optional[Path],
    builtin_templates_name: Optional[str],
) -> TemplateSet:
    """Load templates.

    If templates_folder is not None, load from this folder.
    Else load embedded templated with specified or default name.
    If the temples_folder is None and builtin_name is not None,
    but default templates folder exists, notify that we ignore it.
    """
    if builtin_templates_name is not None and templates_folder is not None:
        log.error("You can't use both --templates-name and --templates-folder")
        sys.exit(1)
    try:
        if templates_folder is None:
            if books_folder is not None and (books_folder / DEFAULT_TEMPLATES_FOLDER).is_dir():
                if builtin_templates_name is None:
                    return TemplatesLoader().load_folder(books_folder / DEFAULT_TEMPLATES_FOLDER)
                log.info(
                    f"Using embedded templates `{builtin_templates_name}, "
                    f"ignore templates in `{DEFAULT_TEMPLATES_FOLDER}`."
                )
                return TemplatesLoader().load_builtin(builtin_templates_name)
            if builtin_templates_name is None:
                return TemplatesLoader().load_builtin()
        elif not templates_folder.is_absolute():
            if books_folder is None:
                raise ValueError(
                    "if template folder is relative, need books folder to build its full path."
                )
            templates_folder = books_folder / templates_folder
        if builtin_templates_name is not None:
            return TemplatesLoader().load_builtin(builtin_templates_name)
        assert templates_folder is not None  # for mypy
        return TemplatesLoader().load_folder(templates_folder)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(f"Error loading templates: {exc}")
        sys.exit(1)


@click.group(invoke_without_command=True)
@click.pass_context
@click.option(
    "--version",
    "version",
    is_flag=True,
    default=False,
    help="Show version.",
    nargs=1,
)
def main(ctx: click.Context, version: bool) -> None:
    """Create md-files from https://www.goodreads.com/ CSV export.

    For example, you can create nice structure in Obsidian.

    How to create goodreads export see in https://www.goodreads.com/review/import

    They declare it to be removed by August 2022,
    but at least in the 2023 it still works.

    Documentation https://andgineer.github.io/goodreads-export/

    To see help on the commands use `goodreads-export COMMAND --help`.
    For example: `goodreads-export import --help`.
    """
    try:
        if version:
            print(f"{VERSION}")
            sys.exit(0)
        if not ctx.invoked_subcommand:
            click.echo("Error: Missing command.")
            click.echo(main.get_help(ctx))
            sys.exit(1)
    except Exception as exc:  # pylint: disable=broad-except
        print(f"\n{exc}")
        sys.exit(1)


@main.command(name="import")
@BOOKS_FOLDER_OPTION
@VERBOSE_OPTION
@TEMPLATES_FOLDER_OPTION
@BUILTIN_TEMPLATES_NAME_OPTION
@click.option(
    "--in",
    "-i",
    "csv_file",
    default=GOODREAD_EXPORT_FILE_NAME,
    help="""Goodreads export file. By default `goodreads_library_export.csv`.
if you specify just folder it will look for file with this name in that folder.""",
    nargs=1,
)
def import_(
    verbose: bool,
    csv_file: str,
    books_folder: Path,
    templates_folder: Optional[Path],
    builtin_name: Optional[str],
) -> None:
    """Convert goodreads export CSV file to markdown files.

    BOOKS_FOLDER
    Folder where we put result. Do not change existed files except authors merge if necessary.

    See details in https://andgineer.github.io/goodreads-export/
    """
    try:
        log = Log(verbose)

        if os.path.isdir(
            csv_file
        ):  # if folder as csv_file try to find goodreads file in that folder
            csv_file = os.path.join(csv_file, GOODREAD_EXPORT_FILE_NAME)
        if not os.path.isfile(csv_file):
            print(f"Goodreads export file '{csv_file}' not found.")
            sys.exit(1)
        log.start(f"Loading reviews from {csv_file}")
        books = GoodreadsBooks(csv_file)
        print(f" loaded {len(books)} reviews.")
        library = merge_authors(
            log=log,
            books_folder=books_folder,
            templates=load_templates(log, books_folder, templates_folder, builtin_name),
        )
        library.dump(books)
        print(
            f"\nAdded {library.stat.books_added} review files, "
            f"{library.stat.authors_added} author files.",
            f"Renamed {library.stat.authors_renamed} authors.",
        )

    except Exception as exc:  # pylint: disable=broad-except
        print(f"\n{exc}")
        sys.exit(1)


@main.command()
@BOOKS_FOLDER_OPTIONAL_OPTION
@VERBOSE_OPTION
@TEMPLATES_FOLDER_OPTION
@BUILTIN_TEMPLATES_NAME_OPTION
def check(
    verbose: bool,
    books_folder: Optional[Path],
    templates_folder: Optional[Path],
    builtin_name: Optional[str],
) -> None:
    """Check templates consistency with extraction regexes.

    BOOKS_FOLDER
    Optional books' folder. Test loading books from the folder.
    Could load templates from it - see `--templates-folder`.
    """
    try:
        log = Log(verbose)
        templates = load_templates(log, books_folder, templates_folder, builtin_name)
        library = Library(  # to run template checks we do not want changes in fs, so no `folder` argument
            log=log, templates=templates
        )
        library.check_templates()
        log.info("Templates are consistent with extraction regexes.")
        if books_folder is not None:
            load_library(log=log, books_folder=books_folder, templates=templates)
    except Exception as exc:  # pylint: disable=broad-except
        print(f"\n{exc}")
        sys.exit(1)


@main.command()
@BOOKS_FOLDER_OPTION
@VERBOSE_OPTION
@TEMPLATES_FOLDER_OPTION
@BUILTIN_TEMPLATES_NAME_OPTION
def merge(
    verbose: bool,
    books_folder: Path,
    templates_folder: Optional[Path],
    builtin_name: Optional[str],
) -> None:
    """Merge authors in the `BOOKS_FOLDER`.

    Unlike `import` do not need goodreads file.

    See https://andgineer.github.io/goodreads-export/ for details.
    """
    try:
        log = Log(verbose)
        library = merge_authors(
            log=log,
            books_folder=books_folder,
            templates=load_templates(log, books_folder, templates_folder, builtin_name),
        )
        print(
            f"Renamed {library.stat.authors_renamed} authors.",
        )
    except Exception as exc:  # pylint: disable=broad-except
        print(f"\n{exc}")
        sys.exit(1)


@main.command()
@BOOKS_FOLDER_OPTIONAL_OPTION
@VERBOSE_OPTION
@TEMPLATES_FOLDER_OPTION
@BUILTIN_TEMPLATES_NAME_OPTION
def init(
    verbose: bool,
    books_folder: Optional[Path],
    templates_folder: Path,
    builtin_name: Optional[str],
) -> None:
    """Create templates folder in path from `--templates-folder`.

    Copy the templates from built-ins specified in `--builtin-name`.

    See https://andgineer.github.io/goodreads-export/ for details.
    """
    try:
        if books_folder is None and templates_folder is None:
            raise ValueError("You should specify `BOOKS_FOLDER` or `--templates-folder`")
        if books_folder is not None and templates_folder is None:
            templates_folder = Path(DEFAULT_TEMPLATES_FOLDER)
        if templates_folder is not None and not templates_folder.is_absolute():
            if books_folder is None:
                raise ValueError(
                    f"To use relative templates folder `{templates_folder}` is necessary `BOOKS_FOLDER`."
                )
            templates_folder = books_folder / templates_folder
        if templates_folder.exists():
            raise ValueError(f"Templates folder `{templates_folder}` already exists.")
        if builtin_name is None:
            builtin_name = DEFAULT_BUILTIN_TEMPLATE
        log = Log(verbose)
        shutil.copytree(str(TemplatesLoader.builtin_folder(builtin_name)), templates_folder)
        log.info(f"Built-in templates `{builtin_name}` copied to `{templates_folder}`")
    except Exception as exc:  # pylint: disable=broad-except
        print(f"\n{exc}")
        sys.exit(1)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



import codecs
import os
import os.path

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open("requirements.dev.txt") as f:
    tests_requirements = f.read().splitlines()


# Solution from https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path: str) -> str:
    """Read file."""
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    """Parse version from file content."""
    for line in read(rel_path).splitlines():
        if line.startswith("VERSION"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="goodreads-export",
    version=get_version("src/goodreads_export/version.py"),
    author="Andrey Sorokin",
    author_email="andrey@sorokin.engineer",
    description=("Convert goodreads books to markdown files, for example for Obsidian."),
    entry_points={
        "console_scripts": [
            "goodreads-export=goodreads_export.main:main",
        ],
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://andgineer.github.io/goodreads-export/",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=requirements,
    tests_require=tests_requirements,
    # dict union operator 3.9, walrus operator, cached property 3.8
    python_requires=">=3.9",
    keywords="goodreads book markdown obsidian",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=True,
    include_package_data=True,
    data_files=[("", ["LICENSE.txt"])],
)



"""Author's file."""  # pylint: disable=duplicate-code
import urllib.parse
from typing import TYPE_CHECKING, Any, Dict, Optional

from goodreads_export.data_file import DataFile
from goodreads_export.series_file import SeriesList
from goodreads_export.templates import AuthorTemplate

if TYPE_CHECKING:
    from goodreads_export.book_file import BookFile


class AuthorFile(DataFile):
    """Author's file.

    On init extract fields from `content` - override name(s).
    To re-parse the `content` call `parse()`.
    `render()` generate `content` from fields.
    """

    name: str  # primary author name
    names: list[str]
    series: SeriesList
    books: list["BookFile"]

    def __init__(
        self,
        *,
        name: str,
        names: Optional[list[str]] = None,
        series: Optional[SeriesList] = None,
        books: Optional[list["BookFile"]] = None,
        **kwargs: Any,
    ) -> None:  # pylint: disable=unused-argument
        """Extract fields from content."""
        super().__init__(**kwargs)
        self.name = name
        self.names = names or []
        self.series = series or SeriesList()
        self.books = books or []
        self.parse()

    def _get_template(self) -> AuthorTemplate:
        """Template."""
        return self.library.templates.author

    def _get_template_context(self) -> Dict[str, Any]:
        """Template context."""
        return {
            "author": self,
            "urlencode": urllib.parse.urlencode,
        }

    def parse(self) -> None:
        """Parse file content.

        Set self.names to the names in the content and self.name to the first name.

        self.names is [] and no change in self.name if no names found in content.
        """
        if self._content is not None:
            self.names = []
            if regex := self._get_template().names_regexes.choose_regex(self._content):
                self.names = [
                    match[regex.name_group] for match in regex.compiled.finditer(self._content)
                ]
                self.name = self.names[0]  # first name is primary

    def render_body(self) -> str:
        """Render file body."""
        return self._get_template().render_body(self._get_template_context())

    def merge(self, other: "AuthorFile") -> None:
        """Merge other author with this one."""
        for book in other.books:
            book.rename_author(self.name)
        for series in other.series:
            series.rename_author(self.name)
        self.books += other.books
        self.series += other.series
        other.delete_file()

    def check(self) -> bool:
        """Check regexps for the template.

        Create file from fields and after that parse it and compare parsed values with the initial fields
        """
        name = self.name
        self.content = self.render_body()
        is_author_parsed = self.names == [name]
        if not is_author_parsed:
            print(f"Author name {name} is not parsed from content\n{self.content}")
            print(f"using the pattern\n{self._get_template().names_regexes[0].regex}")
        return is_author_parsed
